# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import tempfile
import time

from kolla_kubernetes.tests import base
from kolla_kubernetes.utils import FileUtils
from kolla_kubernetes.utils import JinjaUtils
from kolla_kubernetes.utils import StringUtils
from kolla_kubernetes.utils import TypeUtils
from kolla_kubernetes.utils import YamlUtils


test_yaml = '''
    openstack_auth:
      auth_url: "http://10.10.10.254:35357"
      password: "aILs81T3MFt8jkboBnhYZoSKLGve0XOMmLiQeagX"
      project_name: "admin"
      username: "admin"
    openstack_logging_debug: "False"
    openstack_region_name: "RegionOne"
    openstack_release: "3.0.0"
'''

test_conf1 = '''
    "n1":
      "nn1": "n1_nn1_v"
      "nn2": "{{ n3.nn1 }}"
    "n2": "{{ n1.nn1 }}"
    "n3":
      "nn1": "a {{- n6 }}"
    "n4": "n4_v"
'''

test_conf2 = '''
    "n3":
      "nn1": "n3_nn1_v"
      "nn2": "n3_nn2_v"
    "n4": "n4_v"
    "n5":
      "nn1": "n5_nn1_v"
    "n6": "n6_v"
'''

test_conf12_merged = '''
    "n1":
      "nn1": "n1_nn1_v"
      "nn2": "{{ n3.nn1 }}"
    "n2": "{{ n1.nn1 }}"
    "n3":
      "nn1": "n3_nn1_v"
      "nn2": "n3_nn2_v"
    "n4": "n4_v"
    "n5":
      "nn1": "n5_nn1_v"
    "n6": "n6_v"
'''

test_conf12_merged_rendered = '''
    "n1":
      "nn1": "n1_nn1_v"
      "nn2": "n3_nn1_v"
    "n2": "n1_nn1_v"
    "n3":
      "nn1": "n3_nn1_v"
      "nn2": "n3_nn2_v"
    "n4": "n4_v"
    "n5":
      "nn1": "n5_nn1_v"
    "n6": "n6_v"
'''


class UtilsTestCase(base.BaseTestCase):
    _test_dir = None

    @staticmethod
    def get_test_dir():
        if UtilsTestCase._test_dir is None:
            UtilsTestCase._test_dir = UtilsTestCase._create_test_dir()
        return UtilsTestCase._test_dir

    @staticmethod
    def _create_test_dir():
        dir_prefix = os.path.join(
            "/tmp",
            'kolla-kubernetes-tests_' +
            datetime.datetime.fromtimestamp(
                time.time()).strftime('%Y%m%d%H%M%S') + "_")
        tmp_dir_path = tempfile.mkdtemp(prefix=dir_prefix)
        return tmp_dir_path


class TestFileUtils(UtilsTestCase):

    def test_write_and_read_file(self):
        write_content = "Hello\nWorld\n"
        filename = os.path.join(UtilsTestCase.get_test_dir(),
                                self.__class__.__name__ +
                                "_test_write_and_read_file.txt")
        FileUtils.write_string_to_file(write_content, filename)
        read_content = FileUtils.read_string_from_file(filename)
        self.assertEqual(write_content, read_content)


class TestJinjaUtils(UtilsTestCase):

    def test_merge_configs_and_self_render(self):
        file1 = os.path.join(UtilsTestCase.get_test_dir(),
                             self.__class__.__name__ +
                             "_file1_merge_configs_and_self_render.txt")
        file2 = os.path.join(UtilsTestCase.get_test_dir(),
                             self.__class__.__name__ +
                             "_file2_merge_configs_and_self_render.txt")
        FileUtils.write_string_to_file(test_conf1, file1)
        FileUtils.write_string_to_file(test_conf2, file2)

        d = JinjaUtils.merge_configs_to_dict([file1, file2])
        d2 = YamlUtils.yaml_dict_from_string(test_conf12_merged)
        self.assertEqual(YamlUtils.yaml_dict_to_string(d),
                         YamlUtils.yaml_dict_to_string(d2))

        d3 = JinjaUtils.dict_self_render(d2)
        d4 = YamlUtils.yaml_dict_from_string(test_conf12_merged_rendered)
        self.assertEqual(YamlUtils.yaml_dict_to_string(d3),
                         YamlUtils.yaml_dict_to_string(d4))


class TestStringUtils(UtilsTestCase):

    def test_pad_str(self):
        self.assertEqual(StringUtils.pad_str(" ", 2, ""), "  ")
        self.assertEqual(StringUtils.pad_str(" ", 3, "  "), "     ")
        self.assertEqual(StringUtils.pad_str("  ", 3, " "), "       ")
        self.assertEqual(StringUtils.pad_str("aa", 2, ""), "aaaa")


class TestTypeUtils(base.BaseTestCase):

    scenarios = [
        ('none', dict(text=None, expect=False)),
        ('empty', dict(text='', expect=False)),
        ('junk', dict(text='unlikely', expect=False)),
        ('no', dict(text='no', expect=False)),
        ('yes', dict(text='yes', expect=True)),
        ('0', dict(text='0', expect=False)),
        ('1', dict(text='1', expect=False)),
        ('True', dict(text='True', expect=True)),
        ('False', dict(text='False', expect=False)),
        ('true', dict(text='true', expect=True)),
        ('false', dict(text='false', expect=False)),
        ('shouty', dict(text='TRUE', expect=True)),
    ]

    def test_str_to_bool(self):
        self.assertEqual(self.expect, TypeUtils.str_to_bool(self.text))


class TestYamlUtils(UtilsTestCase):

    def test_write_and_read_string(self):
        dict1 = YamlUtils.yaml_dict_from_string(test_yaml)
        str1 = YamlUtils.yaml_dict_to_string(dict1)
        dict2 = YamlUtils.yaml_dict_from_string(str1)
        str2 = YamlUtils.yaml_dict_to_string(dict2)
        self.assertEqual(str1, str2)

    def test_write_and_read_file(self):
        filename = os.path.join(
            UtilsTestCase.get_test_dir(),
            self.__class__.__name__ +
            "_test_write_and_read_file.txt")
        dict1 = YamlUtils.yaml_dict_from_string(test_yaml)
        YamlUtils.yaml_dict_to_file(dict1, filename)
        dict2 = YamlUtils.yaml_dict_from_file(filename)
        str1 = YamlUtils.yaml_dict_to_string(dict1)
        str2 = YamlUtils.yaml_dict_to_string(dict2)
        self.assertEqual(str1, str2)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from oslo_log import log

from kolla_kubernetes.utils import ExecUtils
from kolla_kubernetes.utils import YamlUtils

LOG = log.getLogger(__name__)


class KubeResourceTypeStatus(object):

    def __init__(self, service_obj, resource_type):

        # Check input args
        if resource_type == 'disk':
            LOG.warn('resource_type disk is not supported yet')
            return

        # Initialize internal vars
        self.service_obj = service_obj
        self.resource_type = resource_type

        self.resource_templates = []

        self.doTemplateAndCheck()

    def asDict(self):
        res = {}
        res['meta'] = {}
        res['meta']['service_name'] = self.service_obj.getName()
        res['meta']['resource_type'] = self.resource_type

        res['results'] = {}
        res['results']['status'] = self.getStatus()

        res['xdetails'] = {}  # add 'x' for sort order and xtra-details

        res['xdetails']['templates'] = []
        for kr in self.resource_templates:
            res['xdetails']['templates'].append(kr.asDict())
        return res

    def getStatus(self):
        for krt in self.resource_templates:
            if krt.getStatus() == 'error':
                return 'error'
        return 'ok'

    def doTemplateAndCheck(self):
        """Checks service resource_type resources in Kubernetes

        For each resourceTemplate of resource_type
          Process the template (which may contain a stream of yaml definitions)
          For each individual yaml definition
            Send to kubernetes
            Compare input definition to output status (do checks!)
            Note: This is kube check only.  Other subcommands should
            take care of application specific health checks (e.g. port checks)
        Summarize all of the above into a results dict
        Prints results dict to stdout as yaml status string
        """

        resourceTemplates = self.service_obj.getResourceTemplatesByType(
            self.resource_type)
        for rt in resourceTemplates:
            file_ = rt.getTemplatePath()

            # Skip unsupported script templates
            if file_.endswith('.sh.j2'):
                LOG.warn('Shell templates are not supported yet. '
                         'Skipping processing status of {}'.format(file_))
                continue

            krt = KubeResourceTemplateStatus(
                self.service_obj, self.resource_type, rt)
            self.resource_templates.append(krt)


class KubeResourceTemplateStatus(object):
    """KubeResourceTemplateStatus

       A KubeResourceTemplateStatus is a jinja template, which when
       processed may generate a stream of KubeResourceYamlStatus definitions
       separated by "^---".  In most cases, the output consists of a
       single KubeResourceYamlStatus blob.  However, sometimes the template
       may print nothing or whitespace (NO-OP).
    """

    def __init__(self, service_obj, resource_type,
                 resource_template_obj):

        # Initialize internal vars
        self.service_obj = service_obj
        self.resource_type = resource_type
        self.resource_template_obj = resource_template_obj

        self.errors = []
        self.oks = []
        self.kube_resources = []

        self.doCheck()

    def asDict(self):
        res = {}
        res['meta'] = {}
        res['meta']['template'] = self.resource_template_obj.getTemplatePath()

        res['results'] = {}
        res['results']['status'] = self.getStatus()
        res['results']['errors'] = self.errors
        res['results']['oks'] = self.oks

        res['xdetails'] = {}  # add 'x' for sort order and xtra-details
        res['xdetails']['segments'] = []
        for kr in self.kube_resources:
            res['xdetails']['segments'].append(kr.asDict())
        return res

    def getStatus(self):
        if len(self.errors) != 0:
            return 'error'
        for kr in self.kube_resources:
            if kr.getStatus() != 'ok':
                return 'error'
        return 'ok'

    def doCheck(self):

        # Build the templating command
        cmd = "kolla-kubernetes resource-template {} {} {} {}".format(
            'create', self.service_obj.getName(), self.resource_type,
            self.resource_template_obj.getName())

        # Execute the command to get the processed template output
        template_out, err = ExecUtils.exec_command(cmd)

        # Skip templates which which produce no-ops (100% whitespace)
        #   (e.g. pv templates for AWS should not use persistent
        #   volumes because AWS uses experimental
        #   auto-provisioning)
        if (err is not None):
            self.errors.append(
                'error processing template file: {}'.format(str(err)))
            return
        elif re.match("^\s+$", template_out):
            msg = "template {} produced empty output (NO-OP)".format(
                self.resource_template_obj.getTemplatePath())
            self.oks.append(msg)
            return

        # If the template output produces a stream of yaml documents
        # which are then piped to kubectl, then we will receive a
        # stream of reports separated by "\n\n".  Split on "\n\n" and
        # process each result individually.  The overall result is the
        # merged output.
        definitions = re.split("^---", template_out, re.MULTILINE)
        for definition in definitions:
            kr = KubeResourceYamlStatus(definition)
            self.kube_resources.append(kr)


class KubeResourceYamlStatus(object):
    """Class represents a single Kube resource yaml blob

    Implements functions to send the blob to "kubectl describe -f -",
    and evaluate the returned info text.
    """

    def __init__(self, kube_resource_definition_yaml):
        # Check input args
        assert len(kube_resource_definition_yaml) > 0

        # Initialize internal vars
        self.y = YamlUtils.yaml_dict_from_string(kube_resource_definition_yaml)
        self.definition = kube_resource_definition_yaml

        self.errors = []
        self.oks = []

        self.doDescribeAndCheck()

    def asDict(self):
        res = {}
        res['meta'] = {}
        res['meta']['name'] = self.getName()
        res['meta']['kind'] = self.getKind()
        res['results'] = {}
        res['results']['status'] = self.getStatus()
        res['results']['errors'] = self.errors
        res['results']['oks'] = self.oks
        return res

    def getKind(self):
        if self.y is None:  # this yaml segment may be empty (comments-only)
            return ""
        assert 'kind' in self.y
        return self.y['kind']

    def getName(self):
        if self.y is None:  # this yaml segment may be empty (comments-only)
            return ""
        assert 'metadata' in self.y
        assert 'name' in self.y['metadata']
        return self.y['metadata']['name']

    def getStatus(self):
        if len(self.errors) == 0:
            return 'ok'
        else:
            return 'error'

    def doDescribeAndCheck(self):

        # This yaml segment may be empty (comments-only)
        if self.y is None:
            self.oks.append('Yaml segment is empty of content and perhaps '
                            'only contains comments')
            return  # Allow to succeed

        # Create the command to send this single resource yaml blob to
        # kubectl to query its existence.
        cmd = ('echo \'{}\' | kubectl describe -f -'.format(
            self.definition.replace("'", "'\\''")))  # escape for bash

        out, err = ExecUtils.exec_command(cmd)

        # Check if kubectl returns non-zero exit status
        if err is not None:
            self.errors.append('Either resource does not exist, '
                               'or invalid resource yaml')
            return

        # For all resource types, check the Name to verify existence
        name = KubeResourceYamlStatus._matchSingleLineField('Name', out)
        if name is None or name != self.getName():
            self.errors.append(
                'No resource with name {} exists'.format(self.getName()))
        else:
            self.oks.append(
                'Verified resource with name {} exists'.format(self.getName()))

        # For PersistentVolumes and PersistentVolumeClaims
        if self.getKind() == 'PersistentVolume' or (
            self.getKind() == 'PersistentVolumeClaim'):

            # Verify that the PV/PVC is bound
            status = KubeResourceYamlStatus._matchSingleLineField(
                'Status', out)
            if status is None or status != "Bound":
                self.errors.append("{} not Bound".format(self.getKind()))
            else:
                self.oks.append("{} Bound".format(self.getKind()))

        # For Services
        if self.getKind() == 'Service':
            # Verify the service has an IP
            ip = KubeResourceYamlStatus._matchSingleLineField('IP', out)
            if ip is None or len(ip) == 0:
                self.errors.append("{} has no IP".format(self.getKind()))
            else:
                self.oks.append("{} has IP".format(self.getKind()))

        # For ReplicationControllers
        # Replicas:       1 current / 1 desired
        # Pods Status:    1 Running / 0 Waiting / 0 Succeeded / 0 Failed
        if self.getKind() == 'ReplicationController':

            # Verify the rc has the right number of replicas
            replicas = KubeResourceYamlStatus._matchSingleLineField(
                'Replicas', out)
            if replicas is None:
                self.errors.append(
                    "{} replicas not found".format(self.getKind()))
            else:
                self.oks.append(
                    "{} replicas found".format(self.getKind()))
                replicas_detail = KubeResourceYamlStatus._matchReturnGroups(
                    '^(\d+) current / (\d+) desired', replicas)
                if replicas_detail is not None:
                    current, desired = replicas_detail
                    if current != desired:
                        self.errors.append(
                            "current != desired: {}".format(replicas))
                    else:
                        self.oks.append(
                            "current == desired: {}".format(replicas))

            # Verify the rc has the right number of pod_status
            pod_status = KubeResourceYamlStatus._matchSingleLineField(
                'Pods Status', out)
            if pod_status is None:
                self.errors.append(
                    "{} pod_status not found".format(self.getKind()))
            else:
                self.oks.append(
                    "{} pod_status found".format(self.getKind()))

                pod_status_detail = KubeResourceYamlStatus._matchReturnGroups(
                    '^(\d+) Running / (\d+) Waiting /'
                    ' (\d+) Succeeded / (\d+) Failed', pod_status)
                if pod_status_detail is not None:
                    running, waiting, succeeded, failed = pod_status_detail
                    if (int(running) == 0 or int(waiting) > 0 or (
                            int(failed) > 0)):
                        self.errors.append(
                            "pod_status has errors {}".format(pod_status))
                    else:
                        self.oks.append(
                            "pod_status has no errors: {}".format(pod_status))

    @staticmethod
    def _matchSingleLineField(field_name, haystack):
        """Returns field name's value"""

        # Initial checks
        assert field_name is not None
        if haystack is None:
            return None

        # Execute the Search
        match = re.search('^{}:\s+(?P<MY_VAL>.*)$'.format(field_name),
                          haystack,
                          re.MULTILINE)

        # Check the value
        if match is None:
            return None
        else:
            return match.group('MY_VAL').strip()

    @staticmethod
    def _matchReturnGroups(regex, haystack):
        """Returns all groups matching regex"""

        # Initial checks
        assert regex is not None
        if haystack is None:
            return None

        # Execute the Search
        match = re.search(regex,
                          haystack,
                          re.MULTILINE)

        # Check the value
        if match is None:
            return None
        else:
            return match.groups()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import os.path
import sys

from oslo_config import cfg
from oslotest import base
import six
import testscenarios


# Python 3, thank you for dropping contextlib.nested
if six.PY3:
    @contextlib.contextmanager
    def nested(*contexts):
        with contextlib.ExitStack() as stack:
            yield [stack.enter_context(c) for c in contexts]
else:
    nested = contextlib.nested


class BaseTestCase(testscenarios.WithScenarios,
                   base.BaseTestCase):
    """Test case base class for all unit tests."""

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.addCleanup(cfg.CONF.reset)
        mod_dir = os.path.dirname(sys.modules[__name__].__file__)
        self.project_dir = os.path.abspath(os.path.join(mod_dir, '..', '..'))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys

from kolla_kubernetes import exception
from oslo_config import cfg

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PathFinder(object):

    _find_cache = {}

    @staticmethod
    def find_installed_root():
        # Full installs use this root path to locate ./share/kolla
        # For system, resolves to /usr/local
        # For virtualenv, resolves to /path/to/venv
        return os.path.abspath(os.path.join(os.path.dirname(
            os.path.realpath(sys.argv[0])), '../'))

    @staticmethod
    def find_development_root():
        # Editable installs (aka. Development: pip install --editable .)
        #   use this root path to locate ../kolla
        # For editable, resolves to /path/to/git/repo/kolla-kubernetes
        return os.path.dirname(
            os.path.dirname(os.path.abspath(__file__)))

    @staticmethod
    def find_kolla_dir():
        return PathFinder._find(KOLLA_SEARCH_PATHS, None)

    @staticmethod
    def find_kolla_kubernetes_dir():
        return PathFinder._find(KOLLA_KUBERNETES_SEARCH_PATHS, None)

    @staticmethod
    def find_config_file(filename):
        search_paths = CONFIG_SEARCH_PATHS
        for d in search_paths:
            f = os.path.join(d, filename)
            if os.path.isfile(f):
                return f
        raise exception.KollaFileNotFoundException(
            "Unable to locate file=[{}] in search_paths=[{}]".format(
                filename, ", ".join(search_paths))
        )

    @staticmethod
    def find_config_files(service_name):
        path = PathFinder.find_config_dir(service_name)
        return PathFinder._list_dir_files(path)

    @staticmethod
    def find_config_dir(service_name):
        return PathFinder._find(CONFIG_SEARCH_PATHS, service_name)

    @staticmethod
    def _find(search_paths, dir_or_file):
        # check the cache first
        cache_key = ("-".join(search_paths) + ":" +
                     (dir_or_file if dir_or_file is not None else ""))
        if cache_key in PathFinder._find_cache:
            return PathFinder._find_cache[cache_key]

        # returns the first directory that exists
        for path in search_paths:
            p = path
            if dir_or_file is not None:
                p = os.path.join(path, dir_or_file)
            if os.path.exists(p):
                PathFinder._find_cache[cache_key] = p
                return p
        raise exception.KollaDirNotFoundException(
            "Unable to locate {} directory in search_paths=[{}]".format(
                dir_or_file, ", ".join(search_paths))
        )

    @staticmethod
    def _list_dir_files(path):
        paths = [os.path.join(path, fn) for fn in next(os.walk(path))[2]]
        return paths


# prioritize directories to search for /etc files
CONFIG_SEARCH_PATHS = [
    # Search installation paths first
    '/etc/kolla',
    '/etc/kolla-kubernetes',
    # Then development paths
    os.path.abspath(os.path.join(PathFinder.find_development_root(),
                                 '../kolla/etc/kolla')),
    os.path.abspath(os.path.join(PathFinder.find_development_root(),
                                 './etc/kolla-kubernetes')),
]

# prioritize directories to search for kolla sources
KOLLA_SEARCH_PATHS = [
    # Search installation paths first
    os.path.abspath(os.path.join(PathFinder.find_installed_root(),
                                 './share/kolla')),
    '/usr/share/kolla',
    # Then search development paths
    os.path.abspath(os.path.join(PathFinder.find_development_root(),
                                 '../kolla')),
    os.path.abspath(os.path.join(PathFinder.find_development_root(),
                                 'kolla')),
]

# prioritize directories to search for kolla-kubernetes sources
KOLLA_KUBERNETES_SEARCH_PATHS = [
    # Search installation paths first
    os.path.abspath(os.path.join(PathFinder.find_installed_root(),
                                 './share/kolla-kubernetes')),
    '/usr/share/kolla-kubernetes',
    # Then search development paths
    os.path.abspath(os.path.join(PathFinder.find_development_root())),
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import os
import sys
import time

from oslo_config import cfg
from oslo_log import log as logging

from kolla_kubernetes.kube_service_status import KubeResourceTypeStatus
from kolla_kubernetes.pathfinder import PathFinder
from kolla_kubernetes.utils import ExecUtils
from kolla_kubernetes.utils import JinjaUtils
from kolla_kubernetes.utils import StringUtils
from kolla_kubernetes.utils import YamlUtils

CONF = cfg.CONF
LOG = logging.getLogger()


class KollaKubernetesResources(object):
    _singleton = None

    _jinja_dict_cache = {}

    @staticmethod
    def Get():
        if KollaKubernetesResources._singleton is None:
            KollaKubernetesResources._singleton = KollaKubernetesResources(
                PathFinder.find_config_file('service_resources.yml'))
        return KollaKubernetesResources._singleton

    @staticmethod
    def GetJinjaDict(service_name=None, cli_args={}, debug_regex=None):
        # check the cache first
        cache_key = ((service_name if service_name is not None else "None") +
                     str(cli_args) +
                     (debug_regex if debug_regex is not None else "None"))
        if cache_key in KollaKubernetesResources._jinja_dict_cache:
            return KollaKubernetesResources._jinja_dict_cache[cache_key]

        # Apply basic variables that aren't defined in any config file
        jvars = {'node_config_directory': '',
                 'timestamp': str(time.time())}

        # Add the cli args to the template vars, to be fed into templates
        jvars["kolla_kubernetes"] = {
            "cli": {"args": YamlUtils.yaml_dict_normalize(cli_args)}}

        # Create the prioritized list of config files that need to be
        # merged.  Search method for config files: locks onto the first
        # path where the file exists.  Search method for template files:
        # locks onto the first path that exists, and then expects the file
        # to be there.
        kolla_dir = PathFinder.find_kolla_dir()
        files = [
            PathFinder.find_config_file('kolla-kubernetes.yml'),
            PathFinder.find_config_file('globals.yml'),
            PathFinder.find_config_file('passwords.yml'),
            os.path.join(kolla_dir, 'ansible/group_vars/all.yml')]
        if service_name is not None:
            service_ansible_file = os.path.join(
                kolla_dir, 'ansible/roles', service_name, 'defaults/main.yml')
            if os.path.exists(service_ansible_file):
                files.append(service_ansible_file)
        files.append(os.path.join(kolla_dir,
                                  'ansible/roles/common/defaults/main.yml'))
        # FIXME probably should move this stuff into
        # ansible/roles/common/defaults/main.yml instead.
        files.append(os.path.join(kolla_dir,
                                  'ansible/roles/haproxy/defaults/main.yml'))

        # Create the config dict
        x = JinjaUtils.merge_configs_to_dict(
            reversed(files), jvars, debug_regex)

        # Render values containing nested jinja variables
        r = JinjaUtils.dict_self_render(x)

        # Add a self referential link so templates can look up things by name.
        r['global'] = r

        # Update the cache
        KollaKubernetesResources._jinja_dict_cache[cache_key] = r
        return r

    def __init__(self, filename):
        if not os.path.isfile(filename):
            print("configuration file={} not found".format(filename))
            sys.exit(1)

        self.filename = filename
        self.y = YamlUtils.yaml_dict_from_file(filename)
        self.services = collections.OrderedDict()
        for i in self.y['kolla-kubernetes']['services']:
            self.services[i['name']] = Service(i)

    def getServices(self):
        return self.services

    def getServiceByName(self, name):
        r = self.getServices()
        if name not in r:
            print("unable to find service={}", name)
            sys.exit(1)
        return r[name]

    def __str__(self):
        s = self.__class__.__name__
        for k, v in self.getServices().items():
            s += "\n" + StringUtils.pad_str(" ", 2, str(v))
        return s


class Service(object):
    VALID_ACTIONS = 'create delete status'.split(" ")
    VALID_RESOURCE_TYPES = ('configmap secret '
                            'disk pv pvc svc bootstrap pod').split(" ")
    # Keep old logic for LEGACY support of bootstrap, run, and kill commands
    #   Legacy commands did not keep order.  Here, we define order.
    #   Hoping to get rid of the LEGACY commands entirely if people okay.
    #   Otherwise, we wait until Ansible workflow engine.
    #   SVC should really be in bootstrap command, since it is stateful
    #   CONFIGMAP remains listed twice, since that was the old logic.
    LEGACY_BOOTSTRAP_RESOURCES = ('configmap secret '
                                  'disk pv pvc bootstrap').split(" ")
    LEGACY_RUN_RESOURCES = 'configmap svc pod'.split(" ")

    def __init__(self, y):
        self.y = y
        self.pods = collections.OrderedDict()
        for i in self.y['pods']:
            self.pods[i['name']] = Pod(i)
        self.resourceTemplates = {}
        for rt in self.VALID_RESOURCE_TYPES:
            # Initialize instance resourceTemplates hash
            if rt not in self.resourceTemplates:
                self.resourceTemplates[rt] = []
            # Skip empty definitions
            if rt not in self.y['resources']:
                continue
            # Handle definitions
            for i in self.y['resources'][rt]:
                self.resourceTemplates[rt].append(ResourceTemplate(i))

    def getName(self):
        return self.y['name']

    def getPods(self):
        return self.pods

    def getPodByName(self, name):
        r = self.getPods()
        if name not in r:
            print("unable to find pod={}", name)
            sys.exit(1)
        return r[name]

    def getResourceTemplatesByType(self, resource_type):
        assert resource_type in self.resourceTemplates
        return self.resourceTemplates[resource_type]

    def getResourceTemplateByTypeAndName(
            self, resource_type, resource_name):

        # create an inverted hash[name]=resourceTemplate
        resourceTemplates = self.getResourceTemplatesByType(resource_type)
        h = {i.getName(): i for i in resourceTemplates}

        # validate
        if resource_name not in h.keys():
            print("unable to find resource_name={}", resource_name)
            sys.exit(1)

        return h[resource_name]

    def __str__(self):
        s = self.__class__.__name__ + " " + self.getName()
        for k, v in self.getPods().items():
            s += "\n" + StringUtils.pad_str(" ", 2, str(v))
        return s

    def do_apply(self, action, resource_types, dry_run=False):
        """Apply action to resource_types

        Example: service.apply("create", "disk")
        Example: service.apply("create", ["disk", "pv", "pvc"])
        Example: service.apply("delete", "all")
        Example: service.apply("status", "all")

        ACTION: string value of (create|delete|status)
        RESOURCE_TYPES: string value of one resource type, or list of
          string values of many resource types
          (configmap|disk|pv|pvc|svc|bootstrap|pod).
          In addition 'all' is a valid resource type.
        """

        # Check action input arg for code errors
        assert type(action) is str
        assert action in Service.VALID_ACTIONS

        # Handle resource_types as string or list, and the special case 'all'
        if type(resource_types) is str:
            if resource_types == 'all':
                resource_types = Service.VALID_RESOURCE_TYPES
            else:
                resource_types = [resource_types]

        # Check resource_type input arg for code errors
        assert type(resource_types) is list
        for t in resource_types:
            assert t in Service.VALID_RESOURCE_TYPES

        # If action is delete, then delete the resource types in
        # reverse order.
        if action == 'delete':
            resource_types = reversed(resource_types)

        # Execute the action for each resource_type
        for rt in resource_types:
            # Handle status action
            if action == "status":
                if rt == "disk":
                    raise Exception('resource type for disk not supported yet')
                krs = KubeResourceTypeStatus(self, rt)
                print(YamlUtils.yaml_dict_to_string(krs.asDict()))
                continue

            # Handle create and delete action
            if rt == "configmap":
                # Take care of configmap as a special case
                for pod in self.getPods().values():
                    for container in pod.getContainers().values():
                        if action == 'create':
                            container.createConfigMaps()
                        elif action == 'delete':
                            container.deleteConfigMaps()
                        else:
                            raise Exception('Code Error')
            else:
                # Handle all other resource_types as the same
                self._ensureResource(action, rt)

    def _ensureResource(self, action, resource_type):
        # Check input args
        assert action in Service.VALID_ACTIONS
        assert resource_type in Service.VALID_RESOURCE_TYPES
        assert resource_type in self.resourceTemplates

        resourceTemplates = self.resourceTemplates[resource_type]

        # If action is delete, then delete the resourceTemplates in
        # reverse order.
        if action == 'delete':
            resourceTemplates = reversed(resourceTemplates)

        for resourceTemplate in resourceTemplates:
            # Build the command based on if shell script or not. If
            # shell script, pipe to sh.  Else, pipe to kubectl
            cmd = "kolla-kubernetes resource-template {} {} {} {}".format(
                action, self.getName(), resource_type,
                resourceTemplate.getName())
            if resourceTemplate.getTemplatePath().endswith('.sh.j2'):
                cmd += " | sh"
            else:
                cmd += " | kubectl {} -f -".format(action)

            # Execute the command
            ExecUtils.exec_command(cmd)


class Pod(object):

    def __init__(self, y):
        self.y = y
        self.containers = collections.OrderedDict()
        for i in self.y['containers']:
            self.containers[i['name']] = Container(i)

    def getName(self):
        return self.y['name']

    def getContainers(self):
        return self.containers

    def getContainerByName(self, name):
        r = self.getContainers()
        if name not in r:
            print("unable to find container={}", name)
            sys.exit(1)
        return r[name]

    def __str__(self):
        s = self.__class__.__name__ + " " + self.getName()
        for k, v in self.getContainers().items():
            s += "\n" + StringUtils.pad_str(" ", 2, str(v))
        return s


class Container(object):

    def __init__(self, y):
        self.y = y

    def getName(self):
        return self.y['name']

    def __str__(self):
        s = self.__class__.__name__
        s += " " + self.getName()
        return s

    def createConfigMaps(self):
        self._ensureConfigMaps('create')

    def deleteConfigMaps(self):
        self._ensureConfigMaps('delete')

    def _ensureConfigMaps(self, action):
        assert action in Service.VALID_ACTIONS

        cmd = ("kubectl {} configmap {}-configmap".format(
            action, self.getName()))

        # For the create action, add some more arguments
        if action == 'create':
            for f in PathFinder.find_config_files(self.getName()):
                cmd += ' --from-file={}={}'.format(
                    os.path.basename(f).replace("_", "-"), f)

        # Execute the command
        ExecUtils.exec_command(cmd)


class ResourceTemplate(object):

    def __init__(self, y):
        # Checks
        assert 'template' in y, str(y)  # not optional
        assert 'name' in y, str(y)  # not optional
        # Construct
        self.y = y

    def getName(self):
        return self.y['name']

    def getTemplate(self):
        return self.y['template']

    def getVars(self):
        return (self.y['vars']
                if 'vars' in self.y else None)  # optional

    def getTemplatePath(self):
        kkdir = PathFinder.find_kolla_kubernetes_dir()
        path = os.path.join(kkdir, self.getTemplate())
        assert os.path.exists(path)
        return path

    def __str__(self):
        s = self.__class__.__name__
        s += " name[{}]".format(
            self.getName() if self.getName() is not None else "")
        s += " template[{}]".format(self.getTemplate())
        s += " vars[{}]".format(
            self.getVars() if self.getVars() is not None else "")
        return s



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import sys

from oslo_log import log

from kolla_kubernetes.commands.base_command import KollaKubernetesBaseCommand
from kolla_kubernetes.service_resources import KollaKubernetesResources
from kolla_kubernetes.service_resources import Service
from kolla_kubernetes.utils import FileUtils
from kolla_kubernetes.utils import JinjaUtils
from kolla_kubernetes.utils import YamlUtils

LOG = log.getLogger(__name__)
KKR = KollaKubernetesResources.Get()


class Resource(KollaKubernetesBaseCommand):
    """Create, delete, or query status for kolla-kubernetes resources"""

    def get_parser(self, prog_name):
        parser = super(Resource, self).get_parser(prog_name)
        parser.add_argument(
            "action",
            metavar="<action>",
            help=("One of [%s]" % ("|".join(Service.VALID_ACTIONS)))
        )
        parser.add_argument(
            "service_name",
            metavar="<service-name>",
            help=("One of [%s]" % ("|".join(KKR.getServices().keys())))
        )
        parser.add_argument(
            "resource_type",
            metavar="<resource-type>",
            help=("One of [%s]" % ("|".join(Service.VALID_RESOURCE_TYPES)))
        )
        return parser

    def take_action(self, args):
        self.validate_args(args)
        service = KKR.getServiceByName(args.service_name)
        service.do_apply(args.action, args.resource_type)

    def validate_args(self, args):
        if args.action not in Service.VALID_ACTIONS:
            msg = ("action [{}] not in valid actions [{}]".format(
                args.action,
                "|".join(Service.VALID_ACTIONS)))
            raise Exception(msg)
        if args.service_name not in KKR.getServices().keys():
            msg = ("service_name [{}] not in valid service_names [{}]".format(
                args.service_name,
                "|".join(KKR.getServices().keys())))
            raise Exception(msg)
        if args.resource_type not in Service.VALID_RESOURCE_TYPES:
            msg = ("resource_type [{}] not in valid resource_types [{}]"
                   .format(args.resource_type,
                           "|".join(Service.VALID_RESOURCE_TYPES)))
            raise Exception(msg)

        service = KKR.getServiceByName(args.service_name)
        if (args.resource_type != 'configmap') and (
            len(service.getResourceTemplatesByType(args.resource_type)) == 0):
            msg = ("service_name [{}] has no resource"
                   " files defined for type [{}]".format(
                       args.service_name, args.resource_type))
            raise Exception(msg)


class ResourceTemplate(Resource):
    """Jinja process kolla-kubernetes resource template files"""

    # This command adds the CLI params as part of the Jinja vars for processing
    # templates. This is needed because some of the templates will need to know
    # the arguments with which this CLI is called.  For example, some
    # resource-type "disk" templates may reference '{{
    # kolla_kubernetes.cli.args.action }}' to produce output such as "gcloud
    # disk create" or "gcloud disk delete" based on the CLI params.  Most
    # templates will not require this, but it is needed for some.

    def get_parser(self, prog_name):
        parser = super(ResourceTemplate, self).get_parser(prog_name)
        parser.add_argument(
            "resource_name",
            metavar="<resource-name>",
            help=("The unique resource-name under service->resource_type")
        )
        parser.add_argument(
            '--print-jinja-vars',
            action='store_true',
            help=("If this boolean is set, the final jinja vars dict used as"
                  " input for template processing will be printed to stderr.  "
                  " The vars dict is created by merging configuration files "
                  " from several sources before applying the dict to itself.")
        ),
        parser.add_argument(
            '--print-jinja-keys-regex',
            metavar='<print-jinja-keys-regex>',
            type=str,
            default=None,
            help=("If this regex string is set, all matching keys encountered"
                  " during the creation of the jinja vars dict will be printed"
                  " to stderr at each stage of processing.  The vars dict is"
                  " created by merging configuration files from several"
                  " sources before applying the dict to itself.")
        )
        return parser

    def take_action(self, args):
        # Validate input arguments
        self.validate_args(args)
        if args.resource_type == 'configmap':
            msg = ("resource-template subcommand for resource-type {} "
                   "is not yet supported".format(args.resource_type))
            raise Exception(msg)

        service = KKR.getServiceByName(args.service_name)
        rt = service.getResourceTemplateByTypeAndName(
            args.resource_type, args.resource_name)

        variables = KKR.GetJinjaDict(service.getName(), vars(args),
                                     args.print_jinja_keys_regex)

        # Merge the template vars with the jinja vars before processing
        variables['kolla_kubernetes'].update(
            {"template": {"vars": rt.getVars()}})

        # handle the debug option --print-jinja-vars
        if args.print_jinja_vars is True:
            print(YamlUtils.yaml_dict_to_string(variables), file=sys.stderr)

        # process the template
        print(JinjaUtils.render_jinja(
            variables,
            FileUtils.read_string_from_file(rt.getTemplatePath())))


class ResourceMap(KollaKubernetesBaseCommand):
    """List available kolla-kubernetes resources to be created or deleted"""

    # If the operator has any question on what Services have what resources,
    # and what resources reference which resourcefiles (on disk), then this
    # command is helpful.  This command prints the available resources in a
    # tree of Service->ResourceType->ResourceFiles.

    def get_parser(self, prog_name):
        parser = super(ResourceMap, self).get_parser(prog_name)
        parser.add_argument(
            "--resource-type",
            metavar="<resource-type>",
            help=("Filter by one of [%s]" % (
                "|".join(Service.VALID_RESOURCE_TYPES)))
        )
        parser.add_argument(
            "--service-name",
            metavar="<service-name>",
            help=("Filter by one of [%s]" % (
                "|".join(KKR.getServices().keys())))
        )
        return parser

    def take_action(self, args):
        for service_name, s in KKR.getServices().items():

            # Skip specific services if the user has defined a filter
            if (args.service_name is not None) and (
                    args.service_name != service_name):
                continue

            print('service[{}]'.format(s.getName()))

            for t in Service.VALID_RESOURCE_TYPES:
                # Skip specific resource_types if the user has defined a filter
                if args.resource_type is not None and args.resource_type != t:
                    continue

                # Skip special case configmaps, which are not defined in the
                #   config file but instead are loaded from searching the kolla
                #   configs.
                if t == 'configmap':
                    continue

                resourceTemplates = s.getResourceTemplatesByType(t)

                print('  resource_type[{}] num_items[{}]'.format(
                    t, len(resourceTemplates)))

                # Print the resource files
                for rt in s.getResourceTemplatesByType(t):
                    print('    ' + str(rt))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import copy
import jinja2
import os
import re
import subprocess
import sys
import yaml

from oslo_log import log as logging

LOG = logging.getLogger()

# Disable yaml short-form printing of aliases and anchors
yaml.SafeDumper.ignore_aliases = lambda self, data: True


def env(*args, **kwargs):
    for arg in args:
        value = os.environ.get(arg)
        if value:
            return value
    return kwargs.get('default', '')


class ExecUtils(object):

    @staticmethod
    def exec_command(cmd):
        """Executes command and returns tuple of (stdout, errorException)

        Callers should check for errorException == None
        """
        # TODO(modify): modify function to include stderr in output tuple
        cmd = cmd.strip()  # strip whitespace
        try:
            LOG.debug("executing cmd[{}]".format(cmd))
            res = subprocess.check_output(
                cmd, shell=True,
                executable='/bin/bash')
            res = res.strip()  # strip whitespace
            LOG.debug("returned[{}]".format(res))
            return (res, None)
        except subprocess.CalledProcessError as e:
            # Any non-zero exit code will result in a thrown exception
            # The stdout may be accessed with e.output
            # The exit code may be accessed with e.returncode
            return (e.output.rstrip(), e)


class FileUtils(object):

    @staticmethod
    def write_string_to_file(s, file):
        # Allows emit exception in error
        with open(file, "w") as f:
            f.write(s)
            f.close()

    @staticmethod
    def read_string_from_file(file):
        # Allows emit exception in error
        data = ""
        with open(file, "r") as f:
            data = f.read()
            f.close()
        return data


class JinjaUtils(object):

    @staticmethod
    def merge_configs_to_dict(config_files, initial_dict=None,
                              debug_regex=None):
        """Create the jinja2 dict, and resolve nested variables

        Returns a copy of the initial_dict, loaded with values from
        config_files.  Order matters.... later config files take precedence
        over earlier config files.

        debug_regex: A regex string, if defined, will print out any matching
        config keys as well as the configuration file it is read from.  Very
        useful for debugging.
          debug_regex = "mariadb"

        The above will print out every single key that matches mariadb.
        Complex regex is supported, since this is passed to re.match
        """

        # If there is an initial dictionary, merge its values first
        d = {}
        if initial_dict is not None:
            d.update(initial_dict)

        # Add the contents of each of the following ansible files into the
        # dict.
        for file_ in config_files:
            try:
                # Merge the configs
                x = YamlUtils.yaml_dict_from_file(file_)
                d.update(x)

                # Handle debug requests
                if debug_regex is not None:
                    print("FILE {}".format(file_), file=sys.stderr)
                    for k, v in x.items():
                        if re.match(debug_regex, k):
                            print("  {}: {}".format(k, v), file=sys.stderr)
            except Exception as e:
                LOG.warning('Unable to read file %s: %s', file_, e)
                raise e
        return d

    @staticmethod
    def render_jinja(dict_, template_str):
        """Render dict onto jinja template and return the string result"""
        name = 'jvars'
        j2env = jinja2.Environment(
            loader=jinja2.DictLoader({name: template_str}))

        # Do not print type for bools "!!bool" on output
        j2env.filters['bool'] = TypeUtils.str_to_bool
        j2env.filters['min'] = min

        # Add a "raise" keyword for raising exceptions from within jinja
        def jinja_raise(message):
            raise Exception(message)
        j2env.globals['raise'] = jinja_raise

        # Add a keyword for accessing KubeUtils from within jinja
        j2env.globals['KubeUtils'] = KubeUtils

        # Render the template
        rendered_template = j2env.get_template(name).render(dict_)
        return rendered_template + "\n"

    @staticmethod
    def dict_self_render(dict_):
        """Render dict_ values containing nested jinja variables

        Resolve these values by rendering the jinja dict on itself, as many
        times as jinja variables contain other jinja variables.  Stop when the
        rendered output stops changing.
        """
        d = copy.deepcopy(dict_)
        template = None
        for i in range(0, 10):
            template = YamlUtils.yaml_dict_to_string_jinja(d)
            rendered_template = JinjaUtils.render_jinja(d, template)
            d = YamlUtils.yaml_dict_from_string(rendered_template)
            if rendered_template.strip() == template.strip():
                return d
        raise Exception("Unable to fully render jinja variables")


class StringUtils(object):

    @staticmethod
    def pad_str(pad, num, s):
        return re.sub("^", (pad * num), s, 0, re.MULTILINE)


class TypeUtils(object):

    @staticmethod
    def str_to_bool(text):
        if not text:
            return False
        if text.lower() in ['true', 'yes']:
            return True
        return False


class YamlUtils(object):

    @staticmethod
    def yaml_dict_to_string(dict_):
        """Convert dict to string for human output"""
        # Use width=1000000 to prevent wrapping
        return yaml.safe_dump(dict_, default_flow_style=False,
                              width=1000000)

    @staticmethod
    def yaml_dict_to_string_jinja(dict_):
        """Convert dict to string for jinja processing

        Use this only for dict strings that jinja will process.  If
        you use the normal style instead of double-quote style, then
        yaml dump will escape all single quotes (') by doubling them
        up ('').  If jinja is to process the string, it will fail.
        Thus, change the quote style to avoid escaping single quote
        (').
        """
        # Use width=1000000 to prevent wrapping
        # Use double-quote style to prevent escaping of ' to ''
        return yaml.safe_dump(dict_, default_flow_style=False,
                              width=1000000, default_style='"')

    @staticmethod
    def yaml_dict_from_string(string_):
        # Use BaseLoader to keep "True|False" strings as strings
        return yaml.load(string_, Loader=yaml.loader.BaseLoader)

    @staticmethod
    def yaml_dict_normalize(dict_):
        # This is used to flip "True|False" typed values back to
        #   strings in a dict.
        return YamlUtils.yaml_dict_from_string(
            YamlUtils.yaml_dict_to_string(dict_))

    @staticmethod
    def yaml_dict_to_file(dict_, file_):
        s = YamlUtils.yaml_dict_to_string(dict_)
        return FileUtils.write_string_to_file(s, file_)

    @staticmethod
    def yaml_dict_from_file(file):
        s = FileUtils.read_string_from_file(file)
        return YamlUtils.yaml_dict_from_string(s)


class KubeUtils(object):

    @staticmethod
    def get_api_url(context):
        """Executes kubectl config view

        Returns either None or a string with API server url.

        Callers should check for returned string if it is not == None
        """
        res, code = ExecUtils.exec_command('kubectl config current-context')
        if code is not None:
            return ('', code)

        current_context = res

        res, code = ExecUtils.exec_command('kubectl config view')
        if code is not None:
            return ('', code)

        configuration = YamlUtils.yaml_dict_from_string(res)

        for cluster in configuration['clusters']:
            server = cluster['cluster']['server']
            context = cluster['name']
            if context == current_context:
                return server

        return None



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from cliff import app
from cliff import commandmanager
from cliff.help import HelpAction

from kolla_kubernetes.version import version_info

PROJECT = 'kolla_kubernetes'
VERSION = version_info.version_string_with_vcs()

# Silence debug messages from particular modules
logging.getLogger("requests").setLevel(logging.INFO)
logging.getLogger("stevedore.extension").setLevel(logging.INFO)


class KollaKubernetesApp(app.App):
    _singleton = None

    @staticmethod
    def Get():
        if KollaKubernetesApp._singleton is None:
            KollaKubernetesApp._singleton = KollaKubernetesApp()
        return KollaKubernetesApp._singleton

    def __init__(self):
        super(self.__class__, self).__init__(
            description='Kolla-Kubernetes command-line interface',
            version=VERSION,
            command_manager=commandmanager.CommandManager(
                'kolla_kubernetes.cli'),
            deferred_help=True)

    def _print_help(self):
        """Generate the help string using cliff.help.HelpAction."""

        action = HelpAction(None, None, default=self)
        action(self.parser, self.options, None, None)

    def initialize_app(self, argv):
        """Overrides: cliff.app.initialize_app

        The cliff.app.run automatically assumes and starts
        interactive mode if launched with no arguments.  Short
        circuit to disable interactive mode, and print help instead.
        """

        if len(argv) == 0:
            self._print_help()

    def build_option_parser(self, description, version):
        """Parse global cli options

        Overrides: cliff.app.build_option_parser

        This class inherits from the cliff.app.App class.  The app.App
        run method will first parse its own options with the parser
        created in this method.  These parsed options end up in its
        own argparse namespace accessible by self.options.  These are
        considered global options.

        The leftover (aka. remainder) command line options that are
        not recognized by this parser are then passed directly to
        subcommand parser located in each subcommand at
        command.Command.get_parser().  That argparse namepace is then
        handed directly to the command.Command.take_action() method of
        each subcommand.  Subcommands may access global options by
        calling KollaKubernetesApp.Get().get_parsed_options().

        """
        parser = super(KollaKubernetesApp, self).build_option_parser(
            description,
            version)

        parser.add_argument(
            '--kube-context',
            metavar='<kube-context>',
            # TODO(when_we_implement_auto_kube_config):
            #  default=KubeUtils.get_current_context(),
            help=('The kubectl context which to use'),
        )
        return parser

    def get_parsed_options(self):
        """Provide a method to allow access to parsed global options"""
        return self.options


def main(argv=sys.argv[1:]):
    kks = KollaKubernetesApp().Get()
    return kks.run(argv)

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
